from collections import defaultdict
import gzip


def load_ncbi_gene(godb, gene_info_file, gene2go_file, tax_id=None):
    """
    Load NCBI gene data to the GODB structure
    :param godb: GODB data structure
    :param gene_info_file: NCBI gene info file
               (ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/gene_info.gz)
    :param gene2go_file: NCBI goenrichment file
               (ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/goenrichment.gz)
    :return: GODB data structure
    """
    gname = {}
    with gzip.open(gene_info_file, 'rb') as f:
        for line in f:
            line = line.decode().strip()
            if not line.startswith('#'):
                f = line.split('\t')
                if not tax_id or f[0] == tax_id:
                    gname[f[1]] = f[2]
    print('Processing gene2go')
    data = defaultdict(set)
    with gzip.open(gene2go_file, 'rb') as f:
        for line in f:
            line = line.decode().strip()
            if not line.startswith('!') and not line.startswith('#'):
                r = line.split('\t')
                if (not tax_id or r[0] == tax_id) and r[1] in gname:
                    alt_id = godb['alt_id'][godb['alt_id']['alt_id'] == r[2]]
                    if not alt_id.empty:
                        r[2] = alt_id.iloc[0]['term']
                    data.setdefault(r[2], set()).add(gname[r[1]])
    return data


def load_uniprot_goa_gene(godb, uniprot_goa_file, tax_id=None):
    """
    Load Uniprot GOA genes into the GODB data structure
    :param godb: GODB data structure
    :param uniprot_goa_file: Uniprot GOA file
           (ftp://ftp.ebi.ac.uk/pub/databases/GO/goa/UNIPROT/goa_uniprot_all.gaf.gz)
    :param tax_id: Tax id to extract
    :return: GODB data structure
    """
    data = defaultdict(set)
    with gzip.open(uniprot_goa_file, 'rb') as f:
        for line in f:
            line = line.decode().strip()
            if not line.startswith('!') and not line.startswith('#'):
                r = line.split('\t')
                if not tax_id or r[12].split(':')[1] == tax_id:
                    alt_id = godb['alt_id'][godb['alt_id']['alt_id'] == r[4]]
                    if not alt_id.empty:
                        r[4] = alt_id.iloc[0]['term']
                    data.setdefault(r[4], set()).add(r[2])
    return data


def load_tsv_gene(godb, tsv_go_file):
    """
    Load TSV file created with parameters:
    Gene name       GO term accession
    :param godb: GODB data structure
    :param tsv_go_file: TSV input file
    :return: GODB data structure
    """
    data = defaultdict(set)
    with gzip.open(tsv_go_file, 'rb') as f:
        for line in f:
            line = line.decode().strip()
            if not line.startswith('#'):
                f = line.split('\t')
                if len(f) > 2 and 'GO:' in f[1]:
                    alt_id = godb['alt_id'][godb['alt_id']['alt_id'] == f[1]]
                    if not alt_id.empty:
                        f[1] = alt_id.iloc[0]['term']
                    data.setdefault(f[1], set()).add(f[0])
    return data



import pandas
import pickle
import itertools
import networkx as nx

from urllib.request import urlopen


def add_term_to_node(node, line):
    fs = line.split(':', 1)
    if fs[0] not in node:
        node[fs[0]] = fs[1].strip()
    else:
        node[fs[0]] += "|" + fs[1].strip()
    return node


def read_go_node(fin):
    node = {}
    for line in fin:
        line = line.strip()
        if not line:
            break
        else:
            node = add_term_to_node(node, line)
    return node['id'], node


def create_edges(node):
    if 'is_a' in node:
        for a in node['is_a'].split('|'):
            yield (a.split(' ')[0].strip(), node['id'])
    if 'relationship' in node:
        for b in node['relationship'].split('|'):
            if 'part_of' in b:
                yield (b.split(' ')[1].strip(), node['id'])


def parse_go_obo(go_obo_file):
    """
    Parse GO obo file
    :param go_obo_file: GO obo file. Exmaple: http://current.geneontology.org/ontology/go.obo
    :return: An array with GO, GO name and GO space
    """
    with open(go_obo_file, 'r') as fin:
        for line in fin:
            line = line.strip()
            if line == '[Term]':
                id, node = read_go_node(fin)
                if 'is_obsolete' not in node:
                    edges = create_edges(node)
                    yield (id, node), edges


def create_go_graph(go_obo_file):
    """
    Creat the GODB data structure from the GO obo file
    :param go_obo_file: GO obo file. Exmaple: http://current.geneontology.org/ontology/go.obo
    :return: GODB data structure
    """
    go = nx.DiGraph()
    entries = parse_go_obo(go_obo_file)
    nodes, edges = zip(*entries)
    go.add_nodes_from(nodes)
    go.add_edges_from(itertools.chain.from_iterable(edges))

    # Creating graph roots from namespace
    go.graph['roots'] = {data['name']: n for n, data in go._node.items()
                         if data['name'] == data['namespace']}

    # Adding depth to the nodes
    for root in go.graph['roots'].values():
        for n, depth in nx.shortest_path_length(go, root).items():
            node = go.nodes[n]
            node['depth'] = min(depth, node.get('depth', float('inf')))

    alt_id = []
    for n in go.nodes:
        if 'alt_id' in go.nodes[n]:
            for a in go.nodes[n]['alt_id'].split('|'):
                alt_id.append([n, a])
    alt_id = pandas.DataFrame(alt_id, columns=['term', 'alt_id'])

    return {
        'graph': go.reverse(),
        'alt_id': alt_id,
        'M': 0
    }


def update_go_graph(godb, values, attribute):
    """
    Up date the node attribute with the values
    :param godb: GODB data structure
    :param values: defaultdict(set) with GO terms as key and set of genes names as values
    :param attribute: node attribute name
    :return: GODB data structure
    """
    for n in nx.topological_sort(godb['graph']):
        current = godb['graph'].nodes[n].setdefault(attribute, set())
        current.update(values.get(n, set()))
        for p in godb['graph'][n]:
            godb['graph'].nodes[p].setdefault(attribute, set()).update(current)
    terms, nodes = zip(*godb['graph'].nodes(data=True))
    godb['M'] = len({x for n in nodes for x in n['genes']})
    return godb


def load_goenrichdb(source):
    """
    Load the goenrichment database from a local file or from an URL
    :param source: Local file path or URL
    :return: GODB data structure
    """
    print('Loading go enrichment DB from: ' + source)
    if source.startswith('http') or source.startswith('ftp'):
        return pickle.load(urlopen(source))
    else:
        return pickle.load(open(source, "rb"))



import pandas
import numpy as np
import statsmodels.stats.multitest as smm

from scipy.stats import hypergeom
from goenrichment.graph import ancestors


def calculate(godb, query,
              alpha=0.05,
              min_category_depth=4,
              min_category_size=3,
              max_category_size=500):
    """
    Calculate the GO enrichment from a list of gene IDs
    :param df: Pandas dataframe with genes IDs
    :param gene_id_column: Gene ID column on the data frame
    :param geneGoDB: geneGoDB data structure
    :param name_space: GO name space
    :return: Pandas data frame
    """
    try:
        vals = []
        pvalues = []
        N = len(set(query))
        terms, nodes = zip(*godb['graph'].nodes(data=True))
        for node in nodes:
            category = node['genes']
            n = len(category)
            hits = query.intersection(category)
            k = len(hits)
            if k > 0 and \
                    (node.get('depth', 0) >= min_category_depth or
                     len(list(godb['graph'].predecessors(node['id']))) == 0) \
                    and min_category_size <= n <= max_category_size:
                p = hypergeom.sf(k - 1, godb['M'], n, N)
                pvalues.append(p)
                vals.append((node['id'], node['name'], node['namespace'], p, k, n, node['depth']))
        correction = smm.multipletests(pvalues, alpha=alpha, method='fdr_bh')
        df = pandas.DataFrame(vals, columns=['term', 'name', 'namespace', 'p', 'k', 'n', 'depth'])
        df['accepted'] = correction[0]
        df['q'] = correction[1]
        df['-1.0log(q)'] = -1.0 * np.log10(correction[1])
        df = df[['term', 'name', 'namespace', 'depth', 'k',
                 'n', 'p', 'q', '-1.0log(q)', 'accepted']]

        nodes = set()
        index = []
        for i, r in df.sort_values('depth', ascending=False).iterrows():
            if r['term'] not in nodes and r['q'] <= alpha:
                index.append(i)
                nodes.update(ancestors(r['term'], godb['graph']))
        return df.loc[index].sort_values('q', ascending=False)
    except ZeroDivisionError:
        return pandas.DataFrame(columns=['term', 'name', 'namespace',
                                         'depth', 'k', 'n', 'p', 'q',
                                         '-1.0log(q)', 'accepted'])



import unittest

from goenrichment.go import load_goenrichdb


class TestSet(unittest.TestCase):
    def setUp(self):
        self.godb = load_goenrichdb("ftp://ftp.ncbi.nlm.nih.gov/pub/goenrichment/goenrichDB_human.pickle")

    def test_godb_alt_id(self):
        self.assertEqual(len(self.godb['alt_id']), 2375)

    def test_godb_graph(self):
        self.assertEqual(len(self.godb['graph']._node), 45006)

    def test_godb_M(self):
        self.assertEqual(self.godb['M'], 39810)



#!/usr/bin/env python
import argparse
import pickle
import sys

from goenrichment.go import create_go_graph, update_go_graph
from goenrichment.parsers import load_ncbi_gene, load_tsv_gene, load_uniprot_goa_gene


def main():
    parser = argparse.ArgumentParser(
        description='Creates pickle data structure used by "goenrich.py"')
    parser.add_argument('--gene_info', help='NCBI gene_info file', required=False)
    parser.add_argument('--gene2go', help='NCBI gene2go file', required=False)
    parser.add_argument('--tsv',
                        help='TSV file with at least two columns: Gene_name<tab>GO terms',
                        required=False)
    parser.add_argument('--goenrichDB', help='Previous created goenrich pickle file. '
                                             'The new genes will be added to this database',
                        required=False)
    parser.add_argument('--goa_uniprot', help='Uniprot GOA file GAF format',
                        required=False)
    parser.add_argument('--gobo', help='UGO Obo file from Gene Ontology',
                        required=False)
    parser.add_argument('--taxid', help='Process genes for tax id if it is possible',
                        required=False)
    parser.add_argument('-o', help='Pickle output file name', required=True)

    args = parser.parse_args()
    pickle_file = args.o
    tax_id = None
    if args.taxid:
        tax_id = args.taxid

    geneGoDB = None
    if args.gobo:
        print('Creating geneGODB data structure')
        geneGoDB = create_go_graph(args.gobo)

    if args.goenrichDB:
        print('Loading pickle file')
        geneGoDB = pickle.load(open(args.goenrichDB, "rb"))

    if not geneGoDB:
        print('Should use one of the options --gobo or --goenrichDB')
        sys.exit(-1)

    print('There are %d alternative ids' % (len(geneGoDB['alt_id'])))
    print('There are %d GO terms' % (len(geneGoDB['graph'].nodes)))

    if args.gene_info:
        gene_info_file = args.gene_info
    else:
        gene_info_file = None

    if args.gene2go:
        gene2go_file = args.gene2go
    else:
        gene2go_file = None

    if gene_info_file and gene2go_file:
        print('Loaging NCBI gene data')
        values = load_ncbi_gene(geneGoDB, gene_info_file,
                                gene2go_file, tax_id)
        geneGoDB = update_go_graph(geneGoDB, values, 'genes')
        print('There are %d genes' % (geneGoDB['M']))

    if args.tsv:
        print('Loaging TSV gene data')
        values = load_tsv_gene(geneGoDB, args.tsv)
        geneGoDB = update_go_graph(geneGoDB, values, 'genes')
        print('There are %d genes' % (geneGoDB['M']))

    if args.goa_uniprot:
        print('Loaging Uniprot gene data')
        values = load_uniprot_goa_gene(geneGoDB, args.goa_uniprot, tax_id)
        geneGoDB = update_go_graph(geneGoDB, values, 'genes')
        print('There are %d genes' % (geneGoDB['M']))

    print('There are %d alternative ids' % (len(geneGoDB['alt_id'])))
    print('There are %d GO terms' % (len(geneGoDB['graph'].nodes)))
    print('There are %d genes' % (geneGoDB['M']))
    pickle.dump(geneGoDB, open(pickle_file, "wb"))



#!/usr/bin/env python
import argparse

from goenrichment.enrichment import calculate
from goenrichment.go import load_goenrichdb


def main():
    parser = argparse.ArgumentParser(
        description='Calculate GO enrichment from a list '
                    'of genes. Default database organism: human')

    parser.add_argument('-i', help='Input list of gene names', required=True)
    parser.add_argument('-o', help='TSV file with all results', required=True)
    parser.add_argument('--goenrichDB', help='Gene2GO pickle file created with '
                                             '"goenrichDB.py". If not provided '
                                             'the database is loaded from:'
                                             '', required=False)
    parser.add_argument('--min_category_depth', help='Min GO term graph depth '
                                                     'to include in the report. '
                                                     'Default: 4', required=False)
    parser.add_argument('--min_category_size', help='Min number of gene in a GO '
                                                    'term to include in the report. '
                                                    'Default: 3', required=False)
    parser.add_argument('--max_category_size', help='Max number of gene in a GO '
                                                    'term to include in the report. '
                                                    'Default: 500', required=False)
    parser.add_argument('--alpha', help='Alpha value for p-value correction. '
                                        'Default: 0.05', required=False)

    args = parser.parse_args()

    output = args.o
    input = args.i

    goenrichDB = "ftp://ftp.ncbi.nlm.nih.gov/pub/goenrichment/goenrichDB_human.pickle"
    if args.goenrichDB:
        goenrichDB = args.goenrichDB

    godb = load_goenrichdb(goenrichDB)

    alpha = 0.05
    if args.alpha:
        alpha = float(args.alpha)

    min_category_depth = 4
    if args.min_category_depth:
        min_category_depth = int(args.min_category_depth)

    min_category_size = 3
    if args.min_category_size:
        min_category_size = int(args.min_category_size)

    max_category_size = 500
    if args.max_category_size:
        max_category_size = int(args.max_category_size)

    query = set()
    with open(input) as fin:
        for line in fin:
            query.add(line.strip())

    print('There are %d alternative ids in database' % (len(godb['alt_id'])))
    print('There are %d GO terms' % (len(godb['graph'].nodes)))
    print('There are %d genes in database' % (godb['M']))
    print('Query size: %d genes' % len(query))

    df = calculate(godb, query, alpha, min_category_depth, min_category_size, max_category_size)
    print('GO terms with q less than %.2f: %d' % (alpha, len(df[df['q'] <= alpha])))
    df.to_csv(output, sep='\t', index=None)



def ancestors(g, graph):
    """
    Extract ancestors nodes from an starting node
    :param g: starting node name
    :param graph: Graph
    :return: a set with node names
    """
    result = {g}
    for o in graph.successors(g):
        result.update(ancestors(o, graph))
    return result



import os
import os.path

from setuptools import find_packages
from setuptools import setup


def readme():
    with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
        return f.read()

# Set __version__
exec(open('goenrichment/__init__.py').read())

setup(
    name='goenrichment',
    packages=find_packages(),
    data_files=[('', ['README.rst'])],
    use_scm_version=True,
    setup_requires=['wheel', 'setuptools_scm'],
    description='GO enrichment analysis from a list of gene names using a precomputed database',
    long_description=readme(),
    long_description_content_type='text/markdown',
    license='Public Domain',
    author='Vera Alvarez, Roberto',
    author_email='veraalva' '@' 'ncbi.nlm.nih.gov',
    maintainer='Vera Alvarez, Roberto',
    maintainer_email='veraalva' '@' 'ncbi.nlm.nih.gov',
    url='https://gitlab.com/r78v10a07/goenrichment',
    install_requires=['numpy',
                      'scipy',
                      'statsmodels',
                      'pandas',
                      'networkx'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Jupyter',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    keywords='Jupyter NGS',
    project_urls={
        'Documentation': 'https://gitlab.com/r78v10a07/goenrichment/issues',
        'Source': 'https://gitlab.com/r78v10a07/goenrichment',
        'Tracker': 'https://gitlab.com/r78v10a07/goenrichment/issues',
    },
    entry_points={
        'console_scripts': [
            'goenrich = goenrichment.goenrich:main',
            'goenrich_createdb = goenrichment.goenrichDB:main'
        ],
    }
)
